
#ifndef JACQL_AVL_TREE_H_INCLUDED
#define JACQL_AVL_TREE_H_INCLUDED

 /** AVL Tree - JacQL
  *
  *  By JackTex, 2016. All copyright reserved.
  *
  *  No duplicates are allowed, two compare functions are required
  *
  *  AVLtree<T> has the public interface:
  *     AVLtree(int (*compareFunc)(const T&, const T&),
  *             int (*compareAgainstIntFunc)(const unsigned int&, const T&))
  *
  *     treeIterator begin()
  *     treeIterator rear() //One before end() or end() if no elements in Tree
  *     treeIterator end()
  *     treeIterator find(const T val)
  *     treeIterator findID(unsigned int id)
  *     treeIterator bound(int lower, int upper) //Upper & Lower inclusive. Iterator will automatically go to end() once behind upper
  *
  *     selectIterator beginSelection() //Should ALWAYS be valid deletion will automatically rearrange selection
  *     selectIterator endSelection()
  *     selectIterator findSelction(const T val) //Find in the current selection
  *     selectIterator findSelectionID(unsigned int id)
  *
  *     void reduce(bool(*fnct)(const T&), bool full=true) //Create selection? Full=true Otherwise just reduce the selection are
  *     void order(int(*fnct)(const T&, const T&), bool full=true)
  *     void reduce_and_order(bool(*fnct1)(const T&), int(*fnct2)(const T&, const T&), bool full=true)
  *
  *     T lookup(bool(*fnct)(const T&), bool full=true) //Searches and returns first occurrence for which fnct returns true
  *     unsigned int tally(bool (*fnct)(const T&), bool full=true)
  *     bool contains(const T val) //Checks if the tree contains val
  *     bool containsID(unsigned int id) //Same as contains, but with id
  *
  *     bool insert(T val) //If val is already in the tree (by comparison function) it won't be added and returns false
  *     template<class InputIterator>
  *         void insert(InputIterator first, InputIterator last) //Insert each element in the range to the tree
  *     treeIterator deleteKey(const T val) //Returns iterator to the successor of the deleted element
  *     treeIterator deleteKeyID(unsigned int id)
  *     selectIterator deleteKey(selectIterator iter) //Similar like deleteKey but argument and return are selectIterators
  *     T operator[](int i)
  *     bool empty() const
  *     void clear() //Deletes all nodes in the tree
  *     unsigned int size() const
  *
  *     void insertSecondTier(T val)//Increments secondTierCount if exists
  *     void deleteSecondTier(T val)//Deletes only if secondTierCount goes to 0
  *
  *  The [] operator and the findID(unsigned int key) functions are run against an ID which compared via the
  *  compareAgainstIntFunc. Note that all IDs are required to be positive and the ID 0 should be reserved for
  *  invalid IDs.
  *  deleteKey deletes the node associated with val. If T is a pointer type it will also free the allocated memory
  *  that val is pointing to. HOWEVER, to find the node compareFunc is used. Therefore, if the given val allocates
  *  a different part of the memory, it will not be deleted (only the data in the tree is touched, but not the given
  *  argument. If the given argument happens to be the same memory allocation of the node, it will consequently be
  *  deleted!).
  *  If an element is deleted it will also be cut out of the selection and it is safe to continue to use the selection
  *  (apart from iterators which were set on the deleted element). However, newly added elements to the tree are not
  *  automatically added to the selection and it requires a new call to reduce/order with full==true to add new
  *  elements to the selection. This is because selections can have many successive calls to reduce and order to slim
  *  down the selection (i.e. to allow "AND" condiditions). It is not safe, maybe not even wanted (i.e. the selection
  *  is very old and it isn't used anymore. It would be unnecessary overhead to keep it up to date), to call the chain
  *  of function calls each time a new element is added.
  *  The iterators are bidirectional, equality comparable and *, -> dereference-able.
  **/

#include <algorithm>
#include "JacQL_Heap.h"



//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Exception * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQLException_AVLtree : public JacQLException
{
    public:
        JacQLException_AVLtree(const char *description) : JacQLException(description) {}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Selection Heap * * * ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T>
class AVLselectionHeap : public JacQLNode<T>
{
    public:
        AVLselectionHeap<T>(){};
        virtual ~AVLselectionHeap(){};

        void push(JacQLNode<T> *newValue);
        JacQLNode<T> *pull();
        void clear(){heap.clear();this->before=this->after=NULL;}
        bool empty(){return (heap.empty());}

        void setCompareFunction(int (*compareFunc)(const T&, const T&)){heap.setCompareFunction(compareFunc);}
    protected:
        virtual JacQLNode<T> *getAddress(JacQLNode<T> *requester, bool forwardGoing);
    private:
        MinHeap<T> heap;
};

//Methods

template<class T>
void AVLselectionHeap<T>::push(JacQLNode<T> *newValue)
{
    newValue->after = newValue->before = this;
    heap.push(newValue);
}

template<class T>
JacQLNode<T> *AVLselectionHeap<T>::pull()
{
    JacQLNode<T> *n = heap.pull();
    n->before = this->before;
    if(this->before != NULL)
        this->before->after = n;
    if(!heap.empty())
    {
        n->after = this;
        this->before = n;
    }
    else
        n->after = this->before = this->after = NULL;
    return n;
}

template<class T>
JacQLNode<T> *AVLselectionHeap<T>::getAddress(JacQLNode<T> *requester, bool forwardGoing)
{
    if(heap.empty())
        throw JacQLException_AVLtree("JacQLException: Tried to access empty heap!");
    JacQLNode<T> *n, *predecessor;
    do{
        predecessor = this->before;
        n = pull();

        if(predecessor == requester && forwardGoing)
            return n;
        else if(n == requester && !forwardGoing)
            return n->before;
    }while(!heap.empty());
    throw JacQLException_AVLtree("Fatal JacQLException: Invalid Requester!");
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * AVL Tree * * * //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template <class T>
class AVLtree
{
    public:
        AVLtree(int (*compareFunc)(const T&, const T&),
                int (*compareAgainstIntFunc)(const unsigned int&, const T&))
                    :selectionFirstNode(NULL),currentHeap(0),root(NULL),counter(0U),cmpFunc(compareFunc),cmpNFunc(compareAgainstIntFunc){}
        ~AVLtree(void){delete root;}

        //Standard Iterator
        class treeIterator: public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t,JacQLNode<T>*,JacQLNode<T>&>
        {
            private:
                JacQLNode<T> *current;
                int upperBound;
                int lowerBound;
                int (*cmp)(const unsigned int& a, const T& b); //Used for the bound -for ids-. Should be identical with the one of the tree
                                               // < 0 if a < b
                                               // = 0 if a == b
                                               // > 0 if a > b
            public:
                    explicit treeIterator(JacQLNode<T> *start, bool keepStart = false)
                    : current(start), upperBound(-1), lowerBound(-1), cmp(NULL)
                    {
                        if(current!=NULL && keepStart==false)
                        {
                            while(current->left!=NULL)
                                current = current->left;
                        }

                    }
                    explicit treeIterator(JacQLNode<T> *start, int lowBound, int upBound, int(*func)(const unsigned int& a, const T& b))
                    : current(start), upperBound(upBound), lowerBound(lowBound), cmp(func)
                    {
                        if(lowerBound < 0 || upperBound < 0)
                            throw JacQLException_AVLtree("JacQLException: Negative upper or lower bound are not allowed!");
                        if(lowerBound > upperBound)
                            throw JacQLException_AVLtree("JacQLException: lowerBound has to be equal or smaller than upper bound!");
                        while(current!=NULL && (*cmp)(lowerBound, current->val) > 0)
                            (*this)++;
                        if(current!=NULL && (*cmp)(upperBound, current->val) < 0)
                            current = NULL;

                    }
                    treeIterator(const treeIterator& other)
                    {
                        current = other.current;
                        upperBound = other.upperBound;
                        lowerBound = other.lowerBound;
                        cmp = other.cmp;
                    }
                    treeIterator& operator++()
                    {
                        if(current!=NULL)
                        {
                            if(current->right!=NULL)
                            {
                                current = current->right;
                                while(current->left!=NULL)
                                    current = current->left;
                            }
                            else
                            {
                                JacQLNode<T> *step;
                                do
                                {
                                    step = current;
                                    current = current->parent;
                                }
                                while(current!=NULL && current->right == step);
                            }
                        }
                        if(current != NULL && upperBound >= 0 && (*cmp)(upperBound, current->val) < 0)
                            current = NULL;
                        return *this;
                    }
                    treeIterator operator++(int)
                    {
                        treeIterator retval = *this;
                        if(current!=NULL)
                        {
                            if(current->right!=NULL)
                            {
                                current = current->right;
                                while(current->left!=NULL)
                                    current = current->left;
                            }
                            else
                            {
                                JacQLNode<T> *step;
                                do
                                {
                                    step = current;
                                    current = current->parent;
                                }
                                while(current!=NULL && current->right == step);
                            }
                        }
                        if(current != NULL && upperBound >= 0 && (*cmp)(upperBound, current->val) < 0)
                            current = NULL;
                        return retval;
                    }

                    treeIterator& operator--()
                    {
                        if(current!=NULL)
                        {
                            if(current->left!=NULL)
                            {
                                current = current->left;
                                while(current->right!=NULL)
                                    current = current->right;
                            }
                            else
                            {
                                JacQLNode<T> *step;
                                do
                                {
                                    step = current;
                                    current = current->parent;
                                }
                                while(current!=NULL && current->left == step);
                            }
                        }
                        if(current != NULL && lowerBound >= 0 && (*cmp)(lowerBound, current->val) > 0)
                            current = NULL;
                        return *this;
                    }
                    treeIterator operator--(int)
                    {
                        treeIterator retval = *this;
                        if(current!=NULL)
                        {
                            if(current->left!=NULL)
                            {
                                current = current->left;
                                while(current->right!=NULL)
                                    current = current->right;
                            }
                            else
                            {
                                JacQLNode<T> *step;
                                do
                                {
                                    step = current;
                                    current = current->parent;
                                }
                                while(current!=NULL && current->left == step);
                            }
                        }
                        if(current != NULL && lowerBound >= 0 && (*cmp)(lowerBound, current->val) > 0)
                            current = NULL;
                        return retval;
                    }
                    T operator*() const
                    {
                        if(current == NULL)
                            throw JacQLException_AVLtree("Fatal JacQLException: AVL Tree iterator references out of range object!");
                        return current->val;
                    }
                    T *operator->() const {return &(operator*());}
                    bool operator==(treeIterator other) const {return (other.current == current);}
                    bool operator!=(treeIterator other) const {return (other.current != current);}
                    friend AVLtree<T>;
        };
        treeIterator begin(){return treeIterator(root);}
        treeIterator rear(){return treeIterator(rear(root), true);}
        treeIterator end(){return treeIterator(NULL);}
        treeIterator find(const T val){return treeIterator(getNode(root, val, cmpFunc), true);}
        treeIterator findID(unsigned int id){return treeIterator(getNode(root, id, cmpNFunc), true);}
        treeIterator bound(int lower, int upper){return treeIterator(getNode(root, (unsigned int)(lower), cmpNFunc, true), lower, upper, cmpNFunc);}

        //Selection Iterator
        class selectIterator: public std::iterator<std::bidirectional_iterator_tag,T,std::ptrdiff_t,JacQLNode<T>*,JacQLNode<T>&>
        {
            private:
                JacQLNode<T> *current;
            public:
                    explicit selectIterator(JacQLNode<T> *start) : current(start) {}
                    selectIterator (const selectIterator& other){current = other.current;}
                    selectIterator& operator++()
                    {
                        if(current!=NULL)
                            current = current->next();
                        return *this;
                    }
                    selectIterator operator++(int)
                    {
                        selectIterator retval = *this;
                        if(current!=NULL)
                            current = current->next();
                        return retval;
                    }

                    selectIterator& operator--()
                    {
                        if(current!=NULL)
                            current = current->prev();
                        return *this;
                    }
                    selectIterator operator--(int)
                    {
                        selectIterator retval = *this;
                        if(current!=NULL)
                            current = current->prev();
                        return retval;
                    }
                    T operator*() const
                    {
                        if(current == NULL)
                            throw JacQLException_AVLtree("Fatal JacQLException: AVL Tree selection iterator references out of range object!");
                        return current->val;
                    }
                    T *operator->() const {return &(operator*());}
                    bool operator==(selectIterator other) const {return (other.current == current);}
                    bool operator!=(selectIterator other) const {return (other.current != current);}
                    friend AVLtree<T>;
        };
        selectIterator beginSelection(){return selectIterator(selectionFirstNode);}
        selectIterator endSelection(){return selectIterator(NULL);}
        selectIterator findSelection(const T val);
        selectIterator findSelectionID(unsigned int id);

        void reduce(bool(*fnct)(const T&), bool full=true);
        void order(int(*fnct)(const T&, const T&), bool full=true);
        void reduce_and_order(bool(*fnct1)(const T&), int(*fnct2)(const T&, const T&), bool full=true);

        T lookup(bool(*fnct)(const T&), bool full=true);
        unsigned int tally(bool (*fnct)(const T&), bool full=true);
        bool contains(const T val){return (getNode(root, val, cmpFunc) != NULL);}
        bool containsID(unsigned int id){return (getNode(root, id, cmpNFunc) != NULL);}

        bool insert(T val){return insert(val, root);}
        template<class InputIterator>
        void insert(InputIterator first, InputIterator last);
        treeIterator deleteKey(const T val);
        treeIterator deleteKeyID(unsigned int id);
        selectIterator deleteKey(selectIterator iter);
        T operator[](int i);
        bool empty() const {return (root==NULL);}
        void clear() {delete root; root=NULL;}
        unsigned int size() const {return (counter);}

        void insertRelation(T val){insert(val, root, true);}
        void deleteRelation(T val){deleteKey(val, root, cmpFunc, true);}

    private:
        //Selection stuff:
        JacQLNode<T> *selectionFirstNode;
        AVLselectionHeap<T> heaps[2];
        int currentHeap; //Flips between 0 and 1 to indicate the last used heap
                          //Before use flip over!

        JacQLNode<T> *root;
        unsigned int counter; //Counts the number of elements in tree
        int (*cmpFunc)(const T& a, const T& b); //Function used to compare two elements such that result:
                                    // < 0 if a < b
                                    // = 0 if a == b
                                    // > 0 if a > b
        int (*cmpNFunc)(const unsigned int& a, const T& b); //As cmpFunc with the difference that a is just an unsigned int -for ids-
        bool (*reduceFunc)(const T&); //Used for a reduced selection as well as search and tally(!)

        void rotateLeft                 ( JacQLNode<T> *a );
        void rotateRight                ( JacQLNode<T> *a );
        bool insert                     (T val, JacQLNode<T> *n, bool secondTier=false);
        template<class U>
        bool deleteKey                  (const U delKey, JacQLNode<T> *n, int (*func)(const U& a, const T& b), bool secondTier=false);
        JacQLNode<T> *successor         (int direct, JacQLNode<T> *n);
        void cutoutNode                 ( JacQLNode<T> *n );
        void rebalance                  ( JacQLNode<T> *n );
        int height                      ( JacQLNode<T> *n );
        JacQLNode<T> *rear              ( JacQLNode<T> *n );
        template<class U>
        JacQLNode<T> *getNode             ( JacQLNode<T> *n, U val, int (*func)(const U& a, const T& b), bool predecessor = false)
        {
            JacQLNode<T> *child;
            if(n == NULL)
                return NULL;
            int r = (*func)(val, n->val);
            if(r==0)
                return n;
            child = getNode((r<0)?n->left:n->right, val, func, predecessor);
            if(!predecessor)
                return child;
            if(child == NULL) //else
                return n;
            return child;
        }
        void clearSelection(JacQLNode<T> *n = NULL)
        {
            if(n == NULL)
                return;
            JacQLNode<T> *next = n->next();
            n->before = n->after = NULL;
            clearSelection(next);
        }
        template<class InputIterator>
        void reduce(InputIterator start, InputIterator finish)
        {
            JacQLNode<T> *last = NULL;
            InputIterator next(NULL);
            selectionFirstNode = NULL;
            while(start!=finish)
            {
                next = start;
                next++;
                if((*reduceFunc)(*start))
                {
                    start.current->before = last;
                    if(last == NULL)
                        selectionFirstNode = start.current;
                    else
                        last->after = start.current;
                    last = start.current;
                }
                else
                    start.current->before = start.current->after = NULL;
                start = next;
            }
            if(last != NULL)
                last->after = NULL;
        }
        template<class InputIterator>
        void order(InputIterator start, InputIterator finish)
        {
            InputIterator next(NULL);
            selectionFirstNode = NULL;
            while(start!=finish)
            {
                next = start;
                next++;
                heaps[currentHeap].push(start.current);
                start = next;
            }
            if(!heaps[currentHeap].empty())
                selectionFirstNode = heaps[currentHeap].pull();
        }
        template<class InputIterator>
        void reduce_and_order(InputIterator start, InputIterator finish)
        {
            InputIterator next(NULL);
            selectionFirstNode = NULL;
            while(start!=finish)
            {
                next = start;
                next++;
                if((*reduceFunc)(*start))
                    heaps[currentHeap].push(start.current);
                start = next;
            }
            if(!heaps[currentHeap].empty())
                selectionFirstNode = heaps[currentHeap].pull();
        }
        template<class InputIterator>
        T lookup(InputIterator start, InputIterator finish)
        {
            while(start!=finish)
            {
                if((*reduceFunc)(*start))
                    return *start;
                start++;
            }
            throw JacQLException_AVLtree("JacQLExcpetion: No element found during lookup!");
        }
        template<class InputIterator>
        unsigned int tally(InputIterator start, InputIterator finish)
        {
            unsigned int t = 0U;
            while(start!=finish)
            {
                if((*reduceFunc)(*start))
                    t++;
                start++;
            }
            return t;
        }

};

template<class T>
typename AVLtree<T>::selectIterator AVLtree<T>::findSelection(const T val)
{
    JacQLNode<T> *n = getNode(root, val, cmpFunc);
    if(n==NULL || (n->prev() == NULL && n->next() == NULL && selectionFirstNode != n)) //Not found or not in selection
        return selectIterator(NULL);
    return selectIterator(n);
}

template<class T>
typename AVLtree<T>::selectIterator AVLtree<T>::findSelectionID(unsigned int id)
{
    JacQLNode<T> *n = getNode(root, id, cmpNFunc);
    if(n==NULL || (n->prev() == NULL && n->next() == NULL && selectionFirstNode != n)) //Not found or not in selection
        return selectIterator(NULL);
    return selectIterator(n);
}

template<class T>
void AVLtree<T>::reduce(bool(*fnct)(const T&), bool full)
{
    reduceFunc = fnct;
    if(full)
    {
        clearSelection(selectionFirstNode);
        reduce(begin(), end());
    }
    else
        reduce(beginSelection(), endSelection());
}

template<class T>
void AVLtree<T>::order(int(*fnct)(const T&, const T&), bool full)
{
    currentHeap = (currentHeap == 1)?0:1;
    heaps[currentHeap].setCompareFunction(fnct);
    heaps[currentHeap].clear();
    if(full)
    {
        clearSelection(selectionFirstNode);
        order(begin(), end());
    }
    else
        order(beginSelection(), endSelection());
}

template<class T>
void AVLtree<T>::reduce_and_order(bool(*fnct1)(const T&), int(*fnct2)(const T&, const T&), bool full)
{
    reduceFunc = fnct1;
    currentHeap = (currentHeap == 1)?0:1;
    heaps[currentHeap].setCompareFunction(fnct2);
    heaps[currentHeap].clear();
    if(full)
    {
        clearSelection(selectionFirstNode);
        reduce_and_order(begin(), end());
    }
    else
        reduce_and_order(beginSelection(), endSelection());
}

template<class T>
T AVLtree<T>::lookup(bool(*fnct)(const T&), bool full)
{
    reduceFunc = fnct;
    if(full)
        return lookup(begin(), end());
    return lookup(beginSelection(), endSelection());
}

template<class T>
unsigned int AVLtree<T>::tally(bool (*fnct)(const T&), bool full)
{
    reduceFunc = fnct;
    if(full)
        return tally(begin(), end());
    return tally(beginSelection(), endSelection());
}

template<class T>
template<class InputIterator>
void AVLtree<T>::insert(InputIterator first, InputIterator last)
{
    for( ;first != last; first++)
        insert((*first), root);
}

template<class T>
typename AVLtree<T>::treeIterator AVLtree<T>::deleteKey(const T val)
{
    treeIterator iter = find(val);
    if(iter == end())
        return iter;
    treeIterator sucIter = iter;
    sucIter++;
    if(deleteKey(val, root, cmpFunc));
        return sucIter;
    return iter;
}

template<class T>
typename AVLtree<T>::treeIterator AVLtree<T>::deleteKeyID(unsigned int id)
{
    treeIterator iter = findID(id);
    if(iter == end())
        return iter;
    treeIterator sucIter = iter;
    sucIter++;
    if(deleteKey(id, root, cmpNFunc));
        return sucIter;
    return iter;
}

template<class T>
typename AVLtree<T>::selectIterator AVLtree<T>::deleteKey(selectIterator iter)
{
    selectIterator sucIter = iter;
    sucIter++;
    if(deleteKey(*iter, root, cmpFunc));
        return sucIter;
    return iter;
}

template<class T>
T AVLtree<T>::operator[](int i)
{
    if(i<0)
        throw JacQLException_AVLtree("Fatal JacQLException: In AVL Tree, negative keys are not allowed!");
    JacQLNode<T> *n = getNode(root, (unsigned int)(i), cmpNFunc);
    if(n == NULL)
        throw JacQLException_AVLtree("JacQLException: In AVL Tree, element is not in tree!");
    return n->val;
}




/* AVL class definition */
template <class T>
void AVLtree<T>::rebalance(JacQLNode<T> *n)
{
    //Calculate balance
    n->balance = height(n->right) - height(n->left);

    //Reset Quick Height
    n->quickHeight = -1;

    if (n->balance == -2) { //Left heavy
        if (height(n->left->left) >= height(n->left->right))
            rotateRight(n);
        else
        {
            rotateLeft(n->left);
            rotateRight(n);
        }
    }
    else if (n->balance == 2) { //Right heavy
        if (height(n->right->right) >= height(n->right->left))
            rotateLeft(n);
        else
        {
            rotateRight(n->right);
            rotateLeft(n);
        }
    }
}

template <class T>
void AVLtree<T>::rotateLeft(JacQLNode<T> *a)
{
    JacQLNode<T> *b = a->right;
    b->parent = a->parent;
    a->right = b->left;

    if (a->right != NULL)
        a->right->parent = a;

    b->left = a;
    a->parent = b;

    if (b->parent != NULL) {
        if (b->parent->right == a) {
            b->parent->right = b;
        }
        else {
            b->parent->left = b;
        }
    }
    else
        root = b;

    a->quickHeight = b->quickHeight = -1;
}

template <class T>
void AVLtree<T>::rotateRight(JacQLNode<T> *a)
{
    JacQLNode<T> *b = a->left;
    b->parent = a->parent;
    a->left = b->right;

    if (a->left != NULL)
        a->left->parent = a;

    b->right = a;
    a->parent = b;

    if (b->parent != NULL) {
        if (b->parent->right == a) {
            b->parent->right = b;
        }
        else {
            b->parent->left = b;
        }
    }
    else
        root = b;

    a->quickHeight = b->quickHeight = -1;
}

template<class T>
JacQLNode<T> *AVLtree<T>::rear(JacQLNode<T> *n)
{
    if(n == NULL)
        return NULL;
    JacQLNode<T> *temp = rear(n->right);
    if(temp == NULL)
        return n;
    return temp;
}

template <class T>
int AVLtree<T>::height(JacQLNode<T> *n)
{
    if (n == NULL)
        return -1;
    else if(n->quickHeight < 0)
        n->quickHeight = 1 + std::max(height(n->left), height(n->right));
    return n->quickHeight;
}

template <class T>
bool AVLtree<T>::insert(T val, JacQLNode<T> *n, bool secondTier)
{
    if(n == NULL) //Root
    {
        root = new JacQLNode<T>(val, NULL);
        counter++;
        return true;
    }
    int i = (*cmpFunc)(val, n->val);
    if(i == 0)//Duplicate key
    {
        if(secondTier)
            (n->relationCount)++;
        return false;
    }
    JacQLNode<T> *next = (i<0)?n->left:n->right; //val < n-val -> Go left
    if(next == NULL)
    {
        ((i<0)?n->left:n->right) = new JacQLNode<T>(val, n);
        counter++;
    }
    else if(insert(val, next, secondTier) == false)
        return false;
    rebalance(n);
    return true;
}

template<class T>
template<class U>
bool AVLtree<T>::deleteKey(const U delKey, JacQLNode<T> *n, int (*func)(const U& a, const T& b), bool secondTier)
{
    if(n == NULL) //Doesn't exist in this tree
        return false;
    int i = (*func)(delKey, n->val);
    if(i == 0)
    {
        if(secondTier && --(n->relationCount)>=1U) //If secondTier and there are still more relations avoid deletion
            return false;
        int dir = (n->balance != 0)?n->balance:-1;
        JacQLNode<T> *suc = successor(dir, (dir<0)?n->left:n->right);
        if(suc == NULL) //One of the children is NULL
        {
            cutoutNode(n);
            if(selectionFirstNode == n)
                selectionFirstNode = n->next();
            delete n;
            counter--;
            return true;
        }
        else
        {
            suc->parent = n->parent;
            suc->left = n->left;
            suc->right = n->right;
            if(suc->parent == NULL)
                root = suc;
            else
            {
                if(suc->parent->left == n)
                    suc->parent->left = suc;
                else
                    suc->parent->right = suc;
            }
            if(suc->left != NULL)
                suc->left->parent = suc;
            if(suc->right != NULL)
                suc->right->parent = suc;

            n->parent = n->left = n->right = NULL;
            if(selectionFirstNode == n)
                selectionFirstNode = n->next();
            delete n;
            rebalance(suc);
            counter--;
            return true;
        }
    }
    bool temp = deleteKey(delKey, (i<0)?n->left:n->right, func, secondTier);
    if(temp)
        rebalance(n);
    return temp;
}

template<class T>
JacQLNode<T> *AVLtree<T>::successor(int dir, JacQLNode<T> *n)
{
    if(n == NULL)
        return NULL;
    JacQLNode<T> *next = (dir<0)?n->right:n->left;
    if(next == NULL)
    {
        cutoutNode(n);
        return n;
    }
    else
    {
        JacQLNode<T> *temp = successor(dir, next);
        rebalance(n);
        return temp;
    }
}

template<class T>
void AVLtree<T>::cutoutNode(JacQLNode<T> *n)
{
    JacQLNode<T> *child = (n->left != NULL)?n->left:n->right; //The child which isn't NULL
    if(n->parent != NULL)
    {
        if(n->parent->left == n)
            n->parent->left = child;
        else
            n->parent->right = child;
    }
    else //Parent is root
        root = child;

    if(child != NULL)
        child->parent = n->parent;
    n->parent = n->left = n->right = NULL;
}

#endif // JACQL_AVL_TREE_H_INCLUDED
