#ifndef JACQL_ASSOCIATION_H_INCLUDED
#define JACQL_ASSOCIATION_H_INCLUDED

/** This file contains all necessary methods and fields for associative (many-to-many) tables **/

#include "JacQL_Template.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Exception * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQLException_Association : public JacQLException
{
    public:
        JacQLException_Association(const char *desc) : JacQLException(desc) {;}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Template * * * //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T, class FRM, class TO>
class JacQL_Association : public JacQL_Template<T>
{
    public:
        typedef typename JacQL_Template<T>::dataIterator DATAITERATOR;
        typedef typename JacQL_Template<T>::selIterator SELITERATOR;

        JacQL_Association(){}
        JacQL_Association(unsigned int fromId, unsigned int toId)
        {
            setDependency(fromEntry, fromId);
            setDependency(toEntry, toId);
        }
        virtual ~JacQL_Association(){};

        virtual void load(ifstream &inStream){}
        virtual void save(ofstream &outStream){}
        virtual JacQL_Association& operator=(const JacQL_Association&){return *this;}

        virtual const void *getData(const int){return NULL;}

        virtual void setDependency(dependency, unsigned int); //Sets the dependency to another entry

        static void associate(unsigned int frmId, unsigned int toId);
        static void seperate(unsigned int frmId, unsigned int toId);
        static void setup(dependency d, const bool enforceFrom = false, const bool enforceTo = false, const bool weak = true);

        static void loadAll(const char *path, dependency d) throw (InvalidFileException);
        static void deleteEntriesWithDependency(dependency, unsigned int); //Deletes all entries with dependency to the stated id and calls deletion recursively
        static void updateEntriesWithDependency(dependency, unsigned int, unsigned int); //Changes the dependencies of all entries with this one
        static DATAITERATOR deleteEntry(T *element){return deleteEntry(element->getId());}
        static DATAITERATOR deleteEntry(const unsigned int); //Deletes the entry and will call deletion for associated entries
        static SELITERATOR deleteEntry(SELITERATOR iter){SELITERATOR n = iter; n++; deleteEntry((*iter)->getId()); return n;} //Same but gives the next iterator position
        static const unsigned int &add(T *newElement);

    private:
        static void associate();

        static dependency fromEntry;
        static dependency toEntry;

        static bool enforceFrom; //Enforce relation on from or to
        static bool enforceTo;
};



template<class T, class FRM, class TO>
dependency JacQL_Association<T, FRM, TO>::fromEntry;

template<class T, class FRM, class TO>
dependency JacQL_Association<T, FRM, TO>::toEntry;

template<class T, class FRM, class TO>
bool JacQL_Association<T, FRM, TO>::enforceFrom;

template<class T, class FRM, class TO>
bool JacQL_Association<T, FRM, TO>::enforceTo;


template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::setDependency(dependency d, unsigned int d_id)
{
    if(d != fromEntry && d != toEntry) //Has to be one to one
        throw JacQLException_Association("JacQLException: Tried to set a dependency which is not in the association!");
    JacQL_Template<T>::setDependency(d, d_id);
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::associate(unsigned int frmId, unsigned int toId)
{
    FRM *from = FRM::getById(frmId);
    if(from == NULL)
        throw JacQLException_Association("JacQLException: There is no entry with the corresponding from Id!");
    TO *to = TO::getById(toId);
    if(to == NULL)
        throw JacQLException_Association("JacQLException: There is no entry with the corresponding to Id!");

    from->firstTierRelations[toEntry].insertRelation(to);
    to->firstTierRelations[fromEntry].insertRelation(from);
    for(int i = 0; i<D_ENDTAG; i++)
    {
        if(TO::firstTierAssociator[i] != D_ENDTAG)
        {
            for(AVLtree<JacQL_Base *>::treeIterator iter = to->firstTierRelations[i].begin();
                iter != to->firstTierRelations[i].end(); iter++)
            {
                from->secondTierRelations[i].insertRelation(*iter);
                (*iter)->secondTierRelations[fromEntry].insertRelation(from);
            }
        }
    }
    for(int i = 0; i<D_ENDTAG; i++)
    {
        if(FRM::firstTierAssociator[i] != D_ENDTAG)
        {
            for(AVLtree<JacQL_Base *>::treeIterator iter = from->firstTierRelations[i].begin();
                iter != from->firstTierRelations[i].end(); iter++)
            {
                to->secondTierRelations[i].insertRelation(*iter);
                (*iter)->secondTierRelations[toEntry].insertRelation(to);
            }
        }
    }
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::seperate(unsigned int frmId, unsigned int toId)
{
    FRM *from = FRM::getById(frmId);
    if(from == NULL)
        throw JacQLException_Association("JacQLException: There is no entry with the corresponding from Id!");
    TO *to = TO::getById(toId);
    if(to == NULL)
        throw JacQLException_Association("JacQLException: There is no entry with the corresponding to Id!");

    for(int i = 0; i<D_ENDTAG; i++)
    {
        if(TO::firstTierAssociator[i] != D_ENDTAG)
        {
            for(AVLtree<JacQL_Base *>::treeIterator iter = to->firstTierRelations[i].begin();
                iter != to->firstTierRelations[i].end(); iter++)
            {
                from->secondTierRelations[i].deleteRelation(*iter);
                (*iter)->secondTierRelations[fromEntry].deleteRelation(from);
            }
        }
    }
    for(int i = 0; i<D_ENDTAG; i++)
    {
        if(FRM::firstTierAssociator[i] != D_ENDTAG)
        {
            for(AVLtree<JacQL_Base *>::treeIterator iter = from->firstTierRelations[i].begin();
                iter != from->firstTierRelations[i].end(); iter++)
            {
                to->secondTierRelations[i].deleteRelation(*iter);
                (*iter)->secondTierRelations[toEntry].deleteRelation(to);
            }
        }
    }
    from->firstTierRelations[toEntry].deleteRelation(to);
    to->firstTierRelations[fromEntry].deleteRelation(from);
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::setup(dependency d, const bool eFrom, const bool eTo, const bool weak)
{
    JacQL_Template<T>::setup(d, weak);
    fromEntry = FRM::getOwnEntry();
    toEntry = TO::getOwnEntry();
    if(FRM::firstTierAssociator[toEntry] != D_ENDTAG)
        throw JacQLException_Association("JacQLException: A first tier associator has already been set. Make sure there is not more than one association between any two tables and that they have been setup before the association!");
    else
        FRM::firstTierAssociator[toEntry] = d;
    if(TO::firstTierAssociator[fromEntry] != D_ENDTAG)
        throw JacQLException_Association("JacQLException: A first tier associator has already been set. Make sure there is not more than one association between any two tables and that they have been setup before the association!");
    else
        TO::firstTierAssociator[fromEntry] = d;

    enforceFrom = eFrom;
    enforceTo = eTo;
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::loadAll(const char *path, dependency d) throw (InvalidFileException)
{
    try
    {
        JacQL_Template<T>::loadAll(path, d);
    }
    catch(InvalidFileException e)
    {
        //TO and FROM should have been previously loaded. If something went wrong here
        //we should enforce deletion if this is required before passing on the exception
        if(enforceFrom)
            FRM::enforceRelation(toEntry);
        if(enforceTo)
            TO::enforceRelation(fromEntry);
        throw e;
    }
    associate();
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::deleteEntriesWithDependency(dependency d, unsigned int id)
{
    if(d == JacQL_Template<T>::ownEntry)
        return;
    if(d != fromEntry && d != toEntry)
        return;
    for(SELITERATOR iter = JacQL_Template<T>::constrain(true, 1, d, id); iter != JacQL_Template<T>::endSel(); iter++)
        seperate((*iter)->getDependency(fromEntry), (*iter)->getDependency(toEntry));
    JacQL_Template<T>::deleteEntriesWithDependency(d, id);
    if(enforceFrom)
        FRM::enforceRelation(toEntry);
    if(enforceTo)
        TO::enforceRelation(fromEntry);
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::updateEntriesWithDependency(dependency d, unsigned int oldId, unsigned int newId)
{
    if(d == JacQL_Template<T>::ownEntry)
        return;
    if(d != fromEntry && d != toEntry)
        return;
    for(SELITERATOR iter = JacQL_Template<T>::constrain(true, 1, d, oldId); iter != JacQL_Template<T>::endSel(); iter++)
        seperate((*iter)->getDependency(fromEntry), (*iter)->getDependency(toEntry));
    JacQL_Template<T>::updateEntriesWithDependency(d, oldId, newId);
    for(SELITERATOR iter = JacQL_Template<T>::constrain(true, 1, d, newId); iter != JacQL_Template<T>::endSel(); iter++)
        associate((*iter)->getDependency(fromEntry), (*iter)->getDependency(toEntry));
}

template<class T, class FRM, class TO>
typename JacQL_Template<T>::dataIterator JacQL_Association<T, FRM, TO>::deleteEntry(const unsigned int id)
{
    T *victim = JacQL_Template<T>::getById(id);
    seperate(victim->getDependency(fromEntry), victim->getDependency(toEntry));
    typename JacQL_Template<T>::dataIterator iter = JacQL_Template<T>::deleteEntry(id);
    if(enforceFrom)
        FRM::enforceRelation(toEntry);
    if(enforceTo)
        TO::enforceRelation(fromEntry);
    return iter;
}

template<class T, class FRM, class TO>
const unsigned int &JacQL_Association<T, FRM, TO>::add(T *newElement)
{
    if(newElement->getDependency(fromEntry) == 0 || newElement->getDependency(toEntry) == 0)
        throw JacQLException_Association("JacQLException: From or To dependency is not set for new association!");
    T *temp = T::getByConstrain(2, //Ensure unique morphism
                                fromEntry, newElement->getDependency(fromEntry),
                                toEntry, newElement->getDependency(toEntry));
    if(temp == NULL)
    {
        associate(newElement->getDependency(fromEntry), newElement->getDependency(toEntry));
        return JacQL_Template<T>::add(newElement);
    }
    delete newElement;
    return temp->getId();
}

template<class T, class FRM, class TO>
void JacQL_Association<T, FRM, TO>::associate()
{
    for(DATAITERATOR iter = JacQL_Template<T>::begin(); iter != JacQL_Template<T>::end(); iter++)
        associate((*iter)->getDependency(fromEntry), (*iter)->getDependency(toEntry));
}


#endif // JACQL_ASSOCIATION_H_INCLUDED
