#ifndef JACQL_BASE_H
#define JACQL_BASE_H

/** JacQL Base - JacQL
  *
  *  By JackTex, 2016. All copyright reserved.
  *
  *  This base class contains all generic operations on a database entry. Moreover, it contains a couple of static
  *  methods in the protected section to perform generic tasks of higher up templates (JacQL_Template). This
  *  is necessary as it is not possible to have a function pointer to a static method of a template.
  *
  *  JacQL_Base has the public interface:
  *     const unsigned int &getId()
  *     const unsigned int &getDependency(dependency d)
  *
  *     AVLtree<JacQL_Base *> &fTR(dependency d) //First Tier Relation Tree
  *     AVLtree<JacQL_Base *> &sTR(dependency d) //Second Tier Relation Tree
  *
  *     vector<AVLtree<JacQL_Base *> > firstTierRelations; //Public so it can freely modified by Association
  *     vector<AVLtree<JacQL_Base *> > secondTierRelations; //DO NOT TOUCH IT!
  *
  *     template<class U>
  *     static void saveText(U *, ofstream&); //Saves a unicode string to file
  *     template<class U>
  *     static void loadText(U **, ifstream&); //Loads a unicode string from file
  *     void deleteImageFile(const char *directory); //Deletes the image file in the directory associated by this->id
  *
  *     static void clearUpDir(const char *path); //Deletes everything in the path that starts with the prefix
  **/

#include <iostream>
#include <fstream>
#include <cstring>
#include <vector>
#include "JacQL_AVL_Tree.h"
#include "JacQL_Custom.h"
using namespace std;


//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Query Constants * * * ///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

enum QueryType{
    QUERY_INVALID,
    QUERY_STRING,
    QUERY_UNICODE,
    QUERY_NUMBER,
    QUERY_FLOAT,
    QUERY_BOOL
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Exception * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQLException_Base : public JacQLException
{
    public:
        JacQLException_Base(const char *desc) : JacQLException(desc) {;}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Constrainer * * * ///////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

//Used for constrains in tally and reducing
struct Constrainer{
    Constrainer() : num(0), conTable(NULL), conId(NULL) {}

    int num; //Number of conditions -if negative: inverted-
    dependency *conTable;
    unsigned int *conId;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * JacQL_Base * * * ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQL_Base
{
    protected:
        JacQL_Base(unsigned int nId=0)
        : id(nId), firstTierRelations(D_ENDTAG, AVLtree<JacQL_Base *>(compare, compareID)), secondTierRelations(D_ENDTAG, AVLtree<JacQL_Base *>(compare, compareID))
        {
            for(int i=0; i<D_ENDTAG; i++)
                dependencies[i] = 0; //Default no dependencies
        }

        unsigned int id; //Id of the element id==0 means that the element is temporary and not saved in the
            //elementTree
    public:
        virtual ~JacQL_Base(){};
        const unsigned int &getId(){return id;}
        const unsigned int &getDependency(dependency d){return dependencies[d];}

        AVLtree<JacQL_Base *> &fTR(dependency d){return firstTierRelations[d];} //First Tier Relation Tree
        AVLtree<JacQL_Base *> &sTR(dependency d){return secondTierRelations[d];} //Second Tier Relation Tree

        vector<AVLtree<JacQL_Base *> > firstTierRelations; //Public so it can freely modified by Association
        vector<AVLtree<JacQL_Base *> > secondTierRelations; //DO NOT TOUCH IT!

        template<class U>
        static void saveText(U *, ofstream&); //Saves a unicode string to file
        template<class U>
        static void loadText(U **, ifstream&); //Loads a unicode string from file
        void deleteImageFile(const char *directory); //Deletes the image file in the directory associated by this->id

        static void clearUpDir(const char *path, const bool fullClear=false); //Deletes everything in the path that starts with the prefix

    protected:
        virtual void setDependency(dependency, unsigned int); //Sets the dependency to another entry
        virtual void load(ifstream&)=0;
        virtual void save(ofstream&)=0;
        //virtual JacQL_Base& operator=(const JacQL_Base&)=0; //TO BE DEFINED IN TEMPLATE

        static int compare(JacQL_Base *const& first, JacQL_Base *const& second);
        static int compareID(const unsigned int& first, JacQL_Base *const& second);
        static int compareUI(const unsigned int& first, const unsigned int& second);
        static bool applyConstraints(JacQL_Base *const& element);
        static bool applyQuery(JacQL_Base *const& element);
        static int applyCompareQuery(JacQL_Base *const& first, JacQL_Base *const& second);
        static bool fullSelect(JacQL_Base *const& element){return true;}

        virtual void saveRequest() = 0; //Place a save request into the AVL Tree -if changes have occurred-
        virtual const Constrainer& getConstraints() = 0;

        //Query Language methods
        virtual const QueryType& getQueryType() = 0;
        virtual const int& getQueryId() = 0;
        virtual const void *getQueryElement() = 0;
        virtual const int& getQueryComparison() = 0;

        virtual const QueryType& getCompareQueryType() = 0;
        virtual const int& getCompareQueryId() = 0;
        virtual const bool getCompareQueryDesc() = 0;

        virtual const void *getData(const int) = 0;

        unsigned int dependencies[D_ENDTAG]; //Direct Dependencies
};

template<class U>
void JacQL_Base::saveText(U *text, ofstream &outStream) //Saves a unicode string to file
{
    size_t length = char_traits<U>::length(text);
    length++;
    outStream.write((char *)&length, sizeof(size_t));
    for(unsigned int i=0; i<length; i++)
        outStream.write((char *)(text+i), sizeof(U));
}

template<class U>
void JacQL_Base::loadText(U **text, ifstream &inStream) //Loads a unicode string from file
{
    size_t length;
    if((*text)!=NULL)
        delete[] (*text);
    inStream.read((char *)&length, sizeof(size_t));
    (*text) = new U[length];
    for(unsigned int i=0; i<length; i++)
        inStream.read((char *)((*text)+i), sizeof(U));
}

#endif // JACQL_BASE_H
