#ifndef JACQL_HEAP_H_INCLUDED
#define JACQL_HEAP_H_INCLUDED

/**
 *
 *  Code based on:
 *  http://www.codeproject.com/Tips/816934/Min-Binary-Heap-Implementation-in-Cplusplus
 ***/

 /** MinHeap - JacQL
  *
  *  By JackTex, 2016. All copyright reserved.
  *
  *  MinHeap()
  *  MinHeap(int (*compareFunc)(const T&, const T&))
  *  virtual ~MinHeap()
  *
  *  void push(JacQLNode<T> *newValue)
  *  void cleanPush(T newValue)//Use strictly clean or not clean!
  *  JacQLNode<T> *pull()
  *  T cleanPull()
  *  void clear() //Empties the heap and sets the sequence number back to 0
  *  bool empty() const
  *  unsigned int size() const
  *  void setCompareFunction(int (*compareFunc)(const T&, const T&))
  *
  *  The Heap gives a STABLE (realised via sequence number) sorting algorithm with the minimum in the root. An external
  *  comparison function is used for comparison int(a, b) with <0 if a<b, =0 if a=b and >0 if a>b. The comparison
  *  function can be changed at any time provided the heap is empty.
  *  All elements must be of type JacQLNode<T> *, however, the JacQLNode<T> is never modified and only used for
  *  comparison of the content (of type T). It makes much more sense to store pointer of the entire JacQLNode<T>,
  *  rather than just elements of type T in the overall context of JacQL.
  **/

#include <vector>
#include "JacQL_Node.h"
#include "JacQL_Exception.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * Exception * * * /////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

class JacQLException_Heap : public JacQLException
{
    public:
        JacQLException_Heap(const char *description) : JacQLException(description){;}
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * HeapNode * * * //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T>
struct HeapNode{
    JacQLNode<T> *nod;
    int seq;
};

//////////////////////////////////////////////////////////////////////////////////////////////////////
// * * * MinHeap * * * ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

template<class T>
class MinHeap
{
    public:
        MinHeap() : nextSeq(0), cmpFunc(NULL){};
        MinHeap(int (*compareFunc)(const T&, const T&)) : nextSeq(0), cmpFunc(compareFunc){};
        virtual ~MinHeap(){}

        void push(JacQLNode<T> *newValue);
        void cleanPush(T newValue);
        JacQLNode<T> *pull();
        T cleanPull();
        void clear(){vec.clear();nextSeq=0;}
        bool empty() const {return (vec.size()==0);}
        unsigned int size() const {return vec.size();}
        void setCompareFunction(int (*compareFunc)(const T&, const T&));
    private:
        std::vector<HeapNode<T> > vec;
        int nextSeq; //Sequence number to make it a stable sort
        int (*cmpFunc)(const T& a, const T& b); //Function used to compare two elements such that result:
                                                // < 0 if a < b
                                                // = 0 if a == b
                                                // > 0 if a > b

        void bubbleDown(int index);
        void bubbleUp(int index);
        int compare(int indexA, int indexB);
};


template<class T>
void MinHeap<T>::bubbleDown(int index)
{
    int length = vec.size();
    int leftChildIndex = 2*index + 1;
    int rightChildIndex = 2*index + 2;

    if(leftChildIndex >= length)
        return; //index is a leaf

    int minIndex = index;

    //if(vec[index] > vec[leftChildIndex])
    if(compare(index, leftChildIndex) > 0)
        minIndex = leftChildIndex;
                                        //vec[minIndex] > vec[rightChildIndex]
    if(rightChildIndex < length && compare(minIndex, rightChildIndex) > 0)
        minIndex = rightChildIndex;

    if(minIndex != index)
    {
        //need to swap
        HeapNode<T> temp = vec[index];
        vec[index] = vec[minIndex];
        vec[minIndex] = temp;
        bubbleDown(minIndex);
    }
}

template<class T>
void MinHeap<T>::bubbleUp(int index)
{
    if(index == 0)
        return;

    int parentIndex = (index-1)/2;

        //vec[parentIndex] > vec[index]
    if(compare(parentIndex, index) > 0)
    {
        HeapNode<T> temp = vec[parentIndex];
        vec[parentIndex] = vec[index];
        vec[index] = temp;
        bubbleUp(parentIndex);
    }
}

template<class T>
void MinHeap<T>::push(JacQLNode<T> *newValue)
{
    int length = vec.size();
    HeapNode<T> node;
    node.nod = newValue;
    node.seq = nextSeq++;
    vec.push_back(node);
    bubbleUp(length);
}

template<class T>
void MinHeap<T>::cleanPush(T newValue)
{
    push(new JacQLNode<T>(newValue, NULL));
}

template<class T>
JacQLNode<T> *MinHeap<T>::pull()
{
    int length = vec.size();
    if(length == 0)
        return NULL;
    JacQLNode<T> *temp = vec[0].nod;

    vec[0] = vec[length-1];
    vec.pop_back();

    bubbleDown(0);
    if(empty())
        nextSeq = 0; //It is safe and therefore a good idea to reset the sequence number if the heap is empty
    return temp;
}

template<class T>
T MinHeap<T>::cleanPull()
{
    JacQLNode<T> *temp = pull();
    T content = temp->val;
    invalidate_or_nop(&temp->val);
    delete temp;
    return content;
}

template<class T>
int MinHeap<T>::compare(int indexA, int indexB)
{
    int c;
    if(cmpFunc == NULL)
        throw JacQLException_Heap("Fatal JacQLException: Tried to use heap without a valid compare function!");
    c = (*cmpFunc)(vec[indexA].nod->val, vec[indexB].nod->val);
    if(c == 0)
        c = vec[indexA].seq - vec[indexB].seq;
    return c;
}

template<class T>
void MinHeap<T>::setCompareFunction(int (*compareFunc)(const T&, const T&))
{
    if(!empty())
        throw JacQLException_Heap("Fatal JacQLException: Tried to change comparison function of heap while heap is non-empty!");
    cmpFunc = compareFunc;
}

#endif // JACQL_HEAP_H_INCLUDED
